package Awale::TextSearch;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);
use Carp;
use DB_File;
use POSIX qw(:errno_h :fcntl_h);

require DynaLoader;

@ISA = qw(DynaLoader);
$VERSION = '0.02';

bootstrap Awale::TextSearch $VERSION;

my $INDEX_EXT=".index";
my $DB_EXT=".db";
my $LOCK_EXT=".lock";

sub DEBUG () {0}

###########################################################
sub new
{
    my($package, $filebase, %p)=@_;
    my $self=bless {filebase=>$filebase, writer=>$p{writer}}, $package;
}

###########################################################
sub files
{
    my($self)=@_;
    return ($self->{filebase}.$INDEX_EXT, $self->{filebase}.$DB_EXT, 
            $self->{filebase}.$LOCK_EXT);
}

###########################################################
sub _lock
{
}

###########################################################
sub _unlock
{
}


###########################################################
sub search
{
    my($self, $string, %P)=@_;

    $self->_lock();

    # memory map the file
    return unless $self->mmap;
    my $rs=RS();

    $string=~s/$rs//g;

    # exact means must end with RS
    $string.=$rs if $P{exact};

    # search for the string, finds a list of offsets in index file
    my @offsets;
    if($P{linear}) {
        DEBUG and warn "Linear search for '$string'";
        @offsets=linear_search($self->{mem_handle}, $string);
    }
    elsif($P{k}) {                # Baeza-Yates-Gonnet search
        $string="$rs$string";
        DEBUG and 
            warn "BYG search '$string'\n";
        @offsets=byg_search($self->{mem_handle}, $string, $P{k});
        $self->_error;
    }
    else {                          # Boyer-Moore search (default)
        $string="$rs$string";

        my $study=bm_study($string);
        
        DEBUG and warn "BM search '$string'\n";
        @offsets=bm_search($self->{mem_handle}, $study);

        # bm_unstudy($study);
    } 

    DEBUG and 0==@offsets and warn "No hits\n";

    return {} unless @offsets;

    return unless $self->open_db;

    # now lookup the offsets in the dbm
    my(%IDs, $array);
    foreach my $I (@offsets) {
        $array=$self->{db}->{$I};
        next unless $array;
        DEBUG and warn "Offset $I: $array\n";
        foreach my $q (split ' ', $array) {
            $IDs{$q}=1;
        }
    }
    return \%IDs;
}

###########################################################
sub insert
{
    my($self, $word, $id)=@_;
    $self->_lock;
    $self->open_db;
    my @offsets;

    $self->mmap();
    my $rs=RS();
    if($self->{mem_handle}) {
        # exact match
        @offsets=linear_search($self->{mem_handle}, "$rs$word$rs");
    }
    # if we couldn't mmap the file, assume it's because it doesn't exist

    if(0==@offsets) {                       # wasn't in index
        $self->munmap();                    # close file
                                            # append to file
        open FILE, ">> $self->{filebase}$INDEX_EXT"
                or die "Unable to create $self->{filebase}.$INDEX_EXT: $!";
        print FILE $rs          if -z "$self->{filebase}$INDEX_EXT";
        my $I=tell FILE;
        print FILE $word.$rs;
        close FILE;
        $self->{db}{$I}=$id;                # put it in the DB
        DEBUG and warn "$word is at offset $I\n";
    }
    else {                                  # was in the index
        warn "Index corrupt!  More then one index for $word"
            if 1 != @offsets;

        $self->{db}{$offsets[0]}.=" $id"    # add new ID to DB
            unless $self->{db}{$offsets[0]} =~ /\b$id\b/;
    }

#    $self->close_db;
}

###########################################################
sub delete
{
    croak "NOT IMPLEMENTED";
}

sub DESTROY
{
    my($self)=@_;
    $self->munmmap;
    $self->close_db;
}


###########################################################
sub open_db
{
    my($self)=@_;
    return 1 if $self->{db};
    my $flags=O_RDONLY;
    $flags=O_CREAT|O_RDWR if $self->{writer};
    my %foo;
    tie %foo, 'DB_File', $self->{filebase}.$DB_EXT, $flags, 0600;
    unless(tied %foo or $! == ENOENT or $self->{writer}) {
        return;
    }
    $self->{db}=\%foo;
    return 1;
}

###########################################################
sub close_db
{
    my($self)=@_;
    return 1 unless $self->{db};
    delete $self->{db};
}

###########################################################
# memory map the file
sub mmap
{
    my($self)=@_;
    return 1 if $self->{mem_handle};

    $self->{mem_handle}=map_file($self->{filebase}.$INDEX_EXT);
    DEBUG and map_dump($self->{mem_handle}) if $self->{mem_handle};
    $self->_error unless $self->{mem_handle};

    return not not $self->{mem_handle};
}

###########################################################
sub munmap
{
    my($self)=@_;
#    return 1 unless $self->{mem_handle};

    delete $self->{mem_handle};
#    unmap_file($q);
#    $self->_error;

}


###########################################################
sub _error
{
    my($self)=@_;
    foreach my $err (search_errors()) {
        $err->{msg}||="Error";
        if($err->{errno}) {
            next if $err->{errno} == ENOENT and $self->{writer};
                    
            $!=$err->{errno};
            $err->{msg}.=": $!";
        }
        carp $err->{msg};
    }
    clear_errors();
}




1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Awale::TextSearch - full text search using a specialised index

=head1 SYNOPSIS

    use Awale::TextSearch;

    # constructor
    my $search=Awale::TextSearch->new('/path/to/basname');

    # searches
    my(%res);
    foreach my $kw (@keywords) {
        foreach my $ID (keys %{$search->search($kw)}) {
            $res{$ID}++;
        }
    }
    
    # insert new words
    foreach my $id (keys %items) {
        foreach my $word (split /\W+/, $items{$id}) {
            $search->insert($word=>$id);
        }
    }


    # delete all files in this db
    foreach my $f ($search->files()) {
        unlink $f;
    }
    

=head1 DESCRIPTION

The purpose of this module is to optimise a full text search of a product
database.  To this end, the text of a product is split into words.  The
words are then put into an index file and associated with a list of item
identifiers in a hash file.  An item identifiers is a short string that
uniquely identifies the item.  It could be SKUs or file names or URLs.


=head2 File format

There are 2 files, an index file and a hash file.

Index file is a "simple" list of words, seperated by a RS (record seperator,
currently space).  The offset of the first letter of each word is an index
(I) which is used to look-up a list in the hash file.

The hash file is an hash file format (currently Berkley DB).  Keys are the
offsets taken from the index file, values are arrays of items identifiers.

Or to put it another way, to find all the items dealing with "windows", we
search the index file for "windows", which will give use the offset of the
"w" in windows (call this I).  We then look up I in the hash file, which
returns list of item IDs.  We can then use the item IDs to pull item
information from... where ever!  Ideally a CDB or something.

Using intersections and unions, you can combine the results of many searches
into a single list of items, ranked by relevance.  This is done in the
SYNOPSIS.

Simplicity of implementation is achieved by memory-mapping the index file.


=head2 File names

The constructor takes a base file name parameter.  3 files are named from
this base name : an index file ($basename.index), a hash file ($basename.db)
and a lock file ($basename.lock).


=head2 Search algorithms

3 search algorithms are implemented : linear search, Boyer-Moore and
Baeza-Yates-Gonnet.  Linear search is a base-line and is only used for
testing purposes.  Boyer-Moore is 3 times faster and is the default. 
Baeza-Yates-Gonnet is much slower then the linear search, but allows for a
certain numbers of errors in the search results.  This might seem like a
problem, but fuzzy searches are often desirable.

There are 2 types of searches : exact and stem searches.  An exact match
means the index is searched for the entire word.  A stem search means the
index is searched for all words that begin with the search term.  This means
that a search for "win" will return "windows", "window", "winner" and so on. 
It goes without saying that doing a stem search for "a" will return one
helluva lot of items. :)


=head2 Inserting

The insert operation are be slow.  The text of an item is split into words. 
A search in the index is done for each word.  If the word is in the index,
the item's ID is appended to the relevant entry in the hash file.  If the
word isn't in the index, it gets appeneded to the index, and a new entry is
made in the hash file.


=head2 Deleting

Deleting is not currently implemented.


=head1 METHODS

=head2 new

    my $search=Awale::TextSearch->new($basefile, %params);

Creates the search object.  Doesn't (currently) do any actions.  Parameters
are the basename of all files used by the DB and a hash containing extra
details.  Currently, only useful detail is "writer".  If true, you will be
able to insert items into the DB.  


=head2 search

    my $IDs=$search->search($term, %params);

Searches the index file for C<$term>.  Returns a hashref, keys are all the item
identifiers, values are 1.  This is returned as a hashref and not an array
for 2 reasons : the information is already in that format and you are
probably going to combine it with other search results using a hash anyway. 

By default, a Boyer-Moore search for words begining with $term is done.

The hash C<%params> affects how the search is done.

=over 4

=item exact

If true, an exact word (whole word) search is done.  

=item linear

If ture, a linear search is done of the index file.  This is slower then
Boyer-Moore and present only for benchmarking.

=item k

If true, a Baeza-Yates-Gonnet search is done, allowing a maximum of C<k>
mismatches.  Large values of C<k> are a bad idea, unless C<$term> is also
long.  

=back


=head2 insert

    $search->insert($word=>$id);

The insert is described above.  It should be noted that you must create an
object with a C<writer> param to C<new()>.


=head2 files

    my @files_used=$search->files();

Returns a list all the files used by the object.  These files might not
exist (say the lock file) but are still returned.


=head2 mmap

    $search->mmap();

Causes the index file to be mapped into memory.  This is automatically done
by C<search()> anyway.  However, you might want to do this before forking,
so that the child shares the map.

=head2 munmap

    $search->munmap();

Cases the index file to unmapped from memory.  Like, dude.


=head1 BUGS

I strongly suspect there's an uninited memory problem with the Boyer-Moore
search algorithm.  I can't find it though.


=head1 AUTHOR

Philip Gwyn <perler at awale.qc.ca>

=head1 SEE ALSO

perl(1).

=cut
